import os
import sys
import random
import numpy as np
import pygame
import pyqtgraph as pg
from PyQt6.QtWidgets import (QApplication, QMainWindow, QVBoxLayout, QHBoxLayout, 
                             QWidget, QPushButton, QSlider, QLabel, QGraphicsDropShadowEffect)
from PyQt6.QtCore import Qt, QTimer
from PyQt6.QtGui import QColor, QFont
import sounddevice as sd
import time

class AdvancedMP3Player(QMainWindow):
    def __init__(self):
        super().__init__()
        self._disco_start_time = time.time()
        self.current_wave_count = 20  # Initialize the wave count
        self.setup_window()
        self.setup_audio()
        self.setup_visualization()
        self.init_ui()
        self.init_visualizer()
        self.apply_modern_styling()
        
        # Automatically start playing music on startup
        # if self.playlist:
        #     self.play_next_track()
        
    def setup_window(self):
        self.setWindowTitle("Quantum Audio Visualizer")
        self.setGeometry(100, 100, 1000, 700)
        
    def apply_modern_styling(self):
        self.setStyleSheet("""
            QMainWindow {
                background-color: #ffffff;
            }
            QPushButton {
                background-color: #f0f4f8;
                color: #2c3e50;
                border: 1px solid #bdc3c7;
                padding: 10px 15px;
                border-radius: 5px;
                font-weight: bold;
                font-size: 14px;
            }
            QPushButton:hover {
                background-color: #d5dbe2;
                border-color: #2980b9;
            }
            QPushButton:pressed {
                background-color: #2980b9;
                color: #ffffff;
            }
        """)
        
        shadow = QGraphicsDropShadowEffect(self)
        shadow.setBlurRadius(15)
        shadow.setColor(QColor(0, 0, 0, 50))
        shadow.setOffset(0, 4)
        self.plot_widget.setGraphicsEffect(shadow)
        
        for btn in [self.play_btn, self.next_btn]:
            btn_shadow = QGraphicsDropShadowEffect(self)
            btn_shadow.setBlurRadius(10)
            btn_shadow.setColor(QColor(0, 0, 0, 30))
            btn_shadow.setOffset(0, 2)
            btn.setGraphicsEffect(btn_shadow)
        
    def setup_audio(self):
        pygame.mixer.init()
        self.playlist = self.get_mp3_files()
        self.current_track = None
        self.is_playing = False
        self.sample_rate = 44100
        self.buffer_size = 2048
        self.audio_stream = None
        # Set initial volume to 70%
        pygame.mixer.music.set_volume(0.7)
        
    def setup_visualization(self):
        self.viz_modes = [
            self.spectrum_analyzer,
            self.wave_form_visualizer,
            self.circular_spectrum,
            self.phase_space_visualizer,
            self.block_spectrum_visualizer,
            self.digital_meter_visualizer,
            self.sound_sensitive_3d_object_visualizer,
            self.artistic_sound_sculpture_visualizer,
            self.letter_field_visualizer,
            self.dynamic_circle_interaction_visualizer,
            self.three_channel_3d_disco_light_sequencer,
            self.spirograph_visualizer,
            self.emoji_visualizer,
            self.rainbow_wave_visualizer
        ]
        self.current_viz_index = 0
        
    def get_mp3_files(self):
        return [f for f in os.listdir('.') if f.endswith('.mp3')]
    
    def init_ui(self):
        main_widget = QWidget()
        main_layout = QVBoxLayout()
        
        self.track_label = QLabel("No Track Selected")
        self.track_label.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.track_label.setStyleSheet("""
            font-size: 16px;
            color: #2c3e50;
            font-weight: bold;
            margin-bottom: 10px;
        """)
        main_layout.addWidget(self.track_label)
        
        self.plot_widget = pg.PlotWidget()
        self.plot_widget.setBackground('white')
        self.plot_widget.setTitle('Audio Visualization', color='#2c3e50')
        self.plot_widget.hideAxis('left')
        main_layout.addWidget(self.plot_widget)
        
        viz_selection_layout = QHBoxLayout()
        viz_names = [
            'Spectrum', 'Waveform', 'Circular', 'Phase Space', 
            'Block Spectrum', 'Digital Meter', '3D Sound',
            'Sound Sculpture', 'Letter Field', 
            'Circle Interaction', 'Disco Light', 'Spirograph', 'Emoji',
            'Rainbow Waves'
        ]
        
        self.viz_buttons = []
        for i, name in enumerate(viz_names):
            btn = QPushButton(name)
            btn.setStyleSheet("""
                QPushButton {
                    background-color: rgb(236, 240, 241);
                    color: rgb(44, 62, 80);
                    border: 1px solid rgb(189, 195, 199);
                    border-radius: 4px;
                    padding: 5px;
                    font-size: 10px;
                }
                QPushButton:hover {
                    background-color: rgb(213, 219, 219);
                }
                QPushButton:pressed {
                    background-color: rgb(192, 198, 204);
                }
            """)
            btn.clicked.connect(lambda checked, idx=i: self.select_visualization(idx))
            viz_selection_layout.addWidget(btn)
            self.viz_buttons.append(btn)
        
        main_layout.addLayout(viz_selection_layout)
        
        control_layout = QHBoxLayout()
        
        self.play_btn = QPushButton("Play")
        self.play_btn.clicked.connect(self.toggle_play)
        control_layout.addWidget(self.play_btn)
        
        self.next_btn = QPushButton("Next")
        self.next_btn.clicked.connect(self.play_next_track)
        control_layout.addWidget(self.next_btn)
        
        self.auto_change_btn = QPushButton("Auto Change")
        self.auto_change_btn.clicked.connect(self.toggle_auto_change)
        control_layout.addWidget(self.auto_change_btn)

        # Volume Control Layout
        volume_layout = QVBoxLayout()
        
        self.volume_slider = QSlider(Qt.Orientation.Horizontal)
        self.volume_slider.setRange(0, 100)
        self.volume_slider.setValue(70)  # Set initial volume to 70%
        self.volume_slider.valueChanged.connect(self.adjust_volume)
        self.volume_slider.setStyleSheet("""
            QSlider::groove:horizontal {
                background: #e0e0e0;
                height: 12px;
                border-radius: 6px;
            }
            QSlider::handle:horizontal {
                background: #2980b9;
                width: 24px;
                height: 24px;
                margin: -6px 0;
                border-radius: 12px;
            }
            QSlider::handle:horizontal:hover {
                background: #3498db;
            }
            QSlider::sub-page:horizontal {
                background: #3498db;
                height: 12px;
                border-radius: 6px;
            }
        """)
        
        self.volume_label = QLabel("70%")
        self.volume_label.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.volume_label.setStyleSheet("""
            QLabel {
                color: #2c3e50;
                font-size: 14px;
                font-weight: bold;
                margin-top: 5px;
            }
        """)
        
        volume_layout.addWidget(self.volume_slider)
        volume_layout.addWidget(self.volume_label)
        control_layout.addLayout(volume_layout)
        
        main_layout.addLayout(control_layout)
        
        self.wave_slider = QSlider(Qt.Orientation.Horizontal)
        self.wave_slider.setRange(3, 40)  # Set range from 3 to 40
        self.wave_slider.setValue(20)  # Default value
        self.wave_slider.valueChanged.connect(self.update_wave_count)
        self.wave_slider.setVisible(False)  # Initially hidden
        
        # Create a label for the scale
        self.wave_scale_label = QLabel("Number of Waves: 20")  # Default value
        self.wave_scale_label.setAlignment(Qt.AlignmentFlag.AlignCenter)
        main_layout.addWidget(self.wave_scale_label)  # Add scale label to the main layout
        main_layout.addWidget(self.wave_slider)  # Add slider to the main layout below the visualization
        
        main_widget.setLayout(main_layout)
        self.setCentralWidget(main_widget)

    def select_visualization(self, index):
        if 0 <= index < len(self.viz_modes):
            self.current_viz_index = index
            
            for btn in self.viz_buttons:
                btn.setStyleSheet("""
                    QPushButton {
                        background-color: rgb(236, 240, 241);
                        color: rgb(44, 62, 80);
                        border: 1px solid rgb(189, 195, 199);
                        border-radius: 4px;
                        padding: 5px;
                        font-size: 10px;
                    }
                    QPushButton:hover {
                        background-color: rgb(213, 219, 219);
                    }
                    QPushButton:pressed {
                        background-color: rgb(192, 198, 204);
                    }
                """)
            
            self.viz_buttons[index].setStyleSheet("""
                QPushButton {
                    background-color: rgb(52, 152, 219);
                    color: white;
                    border: 1px solid rgb(41, 128, 185);
                    border-radius: 4px;
                    padding: 5px;
                    font-size: 10px;
                }
            """)
        
        if self.viz_buttons[index].text() == 'Rainbow Waves':
            self.wave_slider.setVisible(True)  # Show slider for rainbow wave visualizer
        else:
            self.wave_slider.setVisible(False)  # Hide slider for other visualizers
        
    def init_visualizer(self):
        self.timer = QTimer()
        self.timer.timeout.connect(self.update_visualization)
        self.timer.start(50)
        
        self.viz_change_timer = QTimer()
        self.viz_change_timer.timeout.connect(self.auto_change_visualization)
        self.auto_change_enabled = False
        
    def toggle_play(self):
        if not self.playlist:
            return
        
        if not self.current_track:
            self.play_next_track()
            return
        
        if self.is_playing:
            pygame.mixer.music.pause()
            self.stop_audio_stream()
            self.play_btn.setText("Play")
            self.is_playing = False
        else:
            pygame.mixer.music.unpause()
            self.start_audio_stream()
            self.play_btn.setText("Pause")
            self.is_playing = True
        
    def play_next_track(self):
        if self.playlist:
            self.stop_audio_stream()
            track = random.choice(self.playlist)
            pygame.mixer.music.load(track)
            pygame.mixer.music.play()
            self.current_track = track
            self.track_label.setText(f"Now Playing: {os.path.basename(track)}")
            self.start_audio_stream()
            self.play_btn.setText("Pause")
            self.is_playing = True
        
    def start_audio_stream(self):
        self.stop_audio_stream()
        self.audio_stream = sd.InputStream(
            samplerate=self.sample_rate, 
            channels=1,
            dtype='float32',
            blocksize=self.buffer_size,
            callback=self.audio_callback
        )
        self.audio_stream.start()
        
    def stop_audio_stream(self):
        if self.audio_stream:
            self.audio_stream.stop()
            self.audio_stream.close()
            self.audio_stream = None
        
    def audio_callback(self, indata, frames, time, status):
        if status:
            print(status)
        self.current_audio_data = indata.flatten()
        
    def adjust_volume(self, value):
        pygame.mixer.music.set_volume(value / 100)
        self.volume_label.setText(f"{value}%")
        
    def auto_change_visualization(self):
        if self.is_playing and self.auto_change_enabled:
            self.current_viz_index = (self.current_viz_index + 1) % len(self.viz_modes)
            
            for btn in self.viz_buttons:
                btn.setStyleSheet("""
                    QPushButton {
                        background-color: rgb(236, 240, 241);
                        color: rgb(44, 62, 80);
                        border: 1px solid rgb(189, 195, 199);
                        border-radius: 4px;
                        padding: 5px;
                        font-size: 10px;
                    }
                    QPushButton:hover {
                        background-color: rgb(213, 219, 219);
                    }
                    QPushButton:pressed {
                        background-color: rgb(192, 198, 204);
                    }
                """)
            
            self.viz_buttons[self.current_viz_index].setStyleSheet("""
                QPushButton {
                    background-color: rgb(52, 152, 219);
                    color: white;
                    border: 1px solid rgb(41, 128, 185);
                    border-radius: 4px;
                    padding: 5px;
                    font-size: 10px;
                }
            """)

    def toggle_auto_change(self):
        self.auto_change_enabled = not self.auto_change_enabled
        
        if self.auto_change_enabled:
            self.viz_change_timer.start(30000)
            self.auto_change_btn.setStyleSheet("""
                QPushButton {
                    background-color: rgb(46, 204, 113);
                    color: white;
                    border: 2px solid rgb(39, 174, 96);
                }
            """)
        else:
            self.viz_change_timer.stop()
            self.auto_change_btn.setStyleSheet("")

    def update_visualization(self):
        if hasattr(self, 'current_audio_data') and self.is_playing:
            self.viz_modes[self.current_viz_index](self.current_audio_data)

    def spectrum_analyzer(self, audio_data):
        spectrum = np.abs(np.fft.rfft(audio_data))
        self.plot_widget.clear()
        
        # Normalize spectrum
        spectrum = spectrum / np.max(spectrum)
        
        self.plot_widget.plot(spectrum, pen='#2980b9')  # Blue tone
        self.plot_widget.setTitle('Spectrum Analyzer', color='#2c3e50')
        
        # Ensure consistent display
        self.plot_widget.getPlotItem().setAspectLocked(False)
        self.plot_widget.getPlotItem().getViewBox().setRange(
            xRange=[0, len(spectrum)], 
            yRange=[0, 1]
        )

    def wave_form_visualizer(self, audio_data):
        self.plot_widget.clear()
        
        # Normalize audio data
        audio_data = audio_data / np.max(np.abs(audio_data))
        
        self.plot_widget.plot(audio_data, pen='#2980b9')  # Blue tone
        self.plot_widget.setTitle('Waveform Visualizer', color='#2c3e50')
        
        # Ensure consistent display
        self.plot_widget.getPlotItem().setAspectLocked(False)
        self.plot_widget.getPlotItem().getViewBox().setRange(
            xRange=[0, len(audio_data)], 
            yRange=[-1, 1]
        )

    def circular_spectrum(self, audio_data):
        spectrum = np.abs(np.fft.rfft(audio_data))
        polar_spectrum = np.abs(np.fft.fft(spectrum))
        
        # Normalize polar spectrum
        polar_spectrum = polar_spectrum / np.max(polar_spectrum)
        
        self.plot_widget.clear()
        self.plot_widget.plot(polar_spectrum, pen='#2980b9')  # Blue tone
        self.plot_widget.setTitle('Circular Spectrum', color='#2c3e50')
        
        # Ensure consistent display
        self.plot_widget.getPlotItem().setAspectLocked(False)
        self.plot_widget.getPlotItem().getViewBox().setRange(
            xRange=[0, len(polar_spectrum)], 
            yRange=[0, 1]
        )

    def phase_space_visualizer(self, audio_data):
        """
        Create a phase space visualization by plotting 
        current sample against next sample (time-delayed plot)
        """
        # Ensure we have enough data
        if len(audio_data) < 2:
            return
        
        # Create phase space plot (time-delayed coordinates)
        x = audio_data[:-1]  # Current samples
        y = audio_data[1:]   # Next samples
        
        # Normalize data
        x = x / np.max(np.abs(x))
        y = y / np.max(np.abs(y))
        
        self.plot_widget.clear()
        self.plot_widget.plot(x, y, 
                               pen=None,  # No connecting lines
                               symbolBrush='#2980b9',  # Blue tone
                               symbolSize=3,  # Small points
                               symbol='o')  # Circular points
        
        self.plot_widget.setTitle('Phase Space Visualization', color='#2c3e50')
        
        # Ensure consistent display
        self.plot_widget.getPlotItem().setAspectLocked(False)
        self.plot_widget.getPlotItem().getViewBox().setRange(
            xRange=[-1, 1], 
            yRange=[-1, 1]
        )

    def block_spectrum_visualizer(self, audio_data):
        """
        Create a block-based spectrum visualization with 23 blocks (bars 2-24)
        using different blue tones and filling the entire visualization space
        """
        # Compute the spectrum
        spectrum = np.abs(np.fft.rfft(audio_data))
        
        # Create 23 frequency bands (bars 2-24)
        num_blocks = 23
        block_size = len(spectrum) // 24  # Divide by total number of potential blocks
        
        # Blue tone gradient
        blue_tones = [
            '#08306b', '#08519c', '#0868ac', '#0880bd', 
            '#2171b5', '#2181c4', '#3498db', '#4dabf7', 
            '#5bc0de', '#6ed3ff', '#7bdcff', '#8ee5ff',
            '#7bdcff', '#6ed3ff', '#5bc0de', '#4dabf7',
            '#3498db', '#2181c4', '#2171b5', '#0880bd',
            '#0868ac', '#08519c', '#08306b'
        ]
        
        self.plot_widget.clear()
        
        # Compute max magnitude for normalization
        max_magnitude = 0
        block_magnitudes = []
        
        # First pass: compute magnitudes for bars 2-24
        for i in range(1, num_blocks + 1):  # Start from 2nd bar
            start = i * block_size
            end = (i + 1) * block_size
            band = spectrum[start:end]
            
            # Compute the average magnitude of the band
            avg_magnitude = np.mean(band) if len(band) > 0 else 0
            block_magnitudes.append(avg_magnitude)
            max_magnitude = max(max_magnitude, avg_magnitude)
        
        # Plot blocks across entire visualization space
        for i, magnitude in enumerate(block_magnitudes):
            # Normalize magnitude to fill entire vertical space
            normalized_magnitude = magnitude / max_magnitude if max_magnitude > 0 else 0
            
            # Plot as a vertical block
            x = [i, i, i+1, i+1]
            y = [0, normalized_magnitude, normalized_magnitude, 0]
            
            # Use corresponding blue tone
            self.plot_widget.plot(x, y, 
                                   fill=True, 
                                   brush=blue_tones[i], 
                                   pen=blue_tones[i])
        
        self.plot_widget.setTitle('Block Spectrum Visualization', color='#2c3e50')
        
        # Ensure the plot fills the entire space
        self.plot_widget.getPlotItem().setAspectLocked(False)
        self.plot_widget.getPlotItem().getViewBox().setRange(
            xRange=[0, num_blocks], 
            yRange=[0, 1]
        )

    def digital_meter_visualizer(self, audio_data):
        """
        Create two digital meters with round displays
        """
        self.plot_widget.clear()
        
        # Compute various audio characteristics
        rms = np.sqrt(np.mean(audio_data**2))  # Root Mean Square
        peak = np.max(np.abs(audio_data))  # Peak amplitude
        
        # Increase sensitivity with logarithmic scaling
        rms_normalized = min(np.log1p(rms * 15), 1)  # More sensitive RMS
        peak_normalized = min(np.log1p(peak * 15), 1)  # More sensitive peak
        
        # Color palette
        bg_color = '#ffffff'  # Light background
        dial_color = '#e6eaf0'  # Slightly darker background for dials
        frame_color = '#bdc3c7'  # Soft gray for frame
        tick_color = '#7f8c8d'  # Darker tick color
        label_color = '#2c3e50'  # Dark blue for labels
        
        def draw_digital_meter(center_x, center_y, value, title, color):
            # Draw a rectangle for the meter
            width = 0.2
            height = 0.8 * value  # Height based on the value
            
            # Draw the meter background
            self.plot_widget.addItem(pg.BarGraphItem(x0=center_x - width/2, y0=center_y, 
                                                      x1=center_x + width/2, y1=center_y + height, 
                                                      pen=color, brush=color))
            
            # Add label with current value
            text_item = pg.TextItem(
                text=f"{title}: {value:.2f}", 
                color=label_color, 
                anchor=(0.5, 0.5)
            )
            text_item.setPos(center_x, center_y + height + 0.05)
            self.plot_widget.addItem(text_item)
        
        # Draw two digital meters
        draw_digital_meter(0.25, 0.3, rms_normalized, 'RMS', '#2980b9')  # Blue for RMS
        draw_digital_meter(0.75, 0.3, peak_normalized, 'Peak', '#e74c3c')  # Red for Peak
        
        # Set title and ensure consistent display
        self.plot_widget.setTitle('Digital Audio Meters', color='#2c3e50')
        self.plot_widget.getPlotItem().setAspectLocked(True, ratio=1)
        self.plot_widget.getPlotItem().getViewBox().setRange(
            xRange=[0, 1], 
            yRange=[0, 1]
        )

    def sound_sensitive_3d_object_visualizer(self, audio_data):
        """
        Create a rotating 3D-like object sensitive to audio characteristics
        """
        self.plot_widget.clear()
        
        # Compute audio characteristics
        rms = np.sqrt(np.mean(audio_data**2))
        peak = np.max(np.abs(audio_data))
        
        # Normalize and enhance sensitivity
        rms_normalized = min(np.log1p(rms * 12), 1)
        peak_normalized = min(np.log1p(peak * 5), 1)
        
        # Color palette
        base_color = '#2980b9'  # Blue base
        accent_color = '#e74c3c'  # Red accent
        
        # Create a rotating polygon-like shape
        def create_rotating_shape(center_x, center_y, size, rotation_factor, color):
            # Number of points in the polygon
            num_points = 6
            
            # Base polygon
            base_angles = np.linspace(0, 2*np.pi, num_points, endpoint=False)
            
            # Deform polygon based on audio characteristics
            deformation = np.sin(base_angles * rotation_factor) * size * rotation_factor
            
            # Create polygon points
            x = center_x + (size + deformation) * np.cos(base_angles)
            y = center_y + (size + deformation) * np.sin(base_angles)
            
            # Plot the polygon
            self.plot_widget.plot(x, y, 
                                   pen={'color': color, 'width': 3},
                                   fill=True, 
                                   brush=pg.mkBrush(color=color, alpha=0.3))
            
            # Add inner lines to create 3D-like effect
            for i in range(num_points):
                inner_x = [center_x, x[i]]
                inner_y = [center_y, y[i]]
                self.plot_widget.plot(inner_x, inner_y, 
                                       pen={'color': color, 'width': 1, 'style': Qt.PenStyle.DotLine})
        
        # Create two interacting shapes
        create_rotating_shape(
            center_x=0.5, 
            center_y=0.5, 
            size=0.3, 
            rotation_factor=rms_normalized * 5, 
            color=base_color
        )
        
        create_rotating_shape(
            center_x=0.5, 
            center_y=0.5, 
            size=0.2, 
            rotation_factor=peak_normalized * 7, 
            color=accent_color
        )
        
        # Set title and ensure consistent display
        self.plot_widget.setTitle('Sound Sensitive 3D Object', color='#2c3e50')
        self.plot_widget.getPlotItem().setAspectLocked(True, ratio=1)
        self.plot_widget.getPlotItem().getViewBox().setRange(
            xRange=[0, 1], 
            yRange=[0, 1]
        )

    def artistic_sound_sculpture_visualizer(self, audio_data):
        """
        Create an artistic, abstract sound sculpture 
        with dynamic, responsive geometry
        """
        self.plot_widget.clear()
        
        # Compute audio characteristics with enhanced sensitivity
        rms = np.sqrt(np.mean(audio_data**2))
        peak = np.max(np.abs(audio_data))
        
        # Logarithmic normalization for more nuanced response
        rms_normalized = min(np.log1p(rms * 20), 1)
        peak_normalized = min(np.log1p(peak * 25), 1)
        
        # Color palette with artistic tones
        base_colors = [
            '#2980b9',   # Deep blue
            '#e74c3c',   # Vibrant red
            '#27ae60',   # Emerald green
            '#f39c12',   # Sunflower yellow
        ]
        
        def create_sound_sculpture(center_x, center_y, base_size, complexity, color):
            # Vary number of points based on audio characteristics
            num_points = int(6 + complexity * 6)  # 6-12 points
            
            # Create base polygon
            base_angles = np.linspace(0, 2*np.pi, num_points, endpoint=False)
            
            # Dynamic deformation parameters
            deformation_intensity = complexity * 0.5
            frequency_variation = complexity * 3
            
            # Create sculptural points
            x = []
            y = []
            for i, angle in enumerate(base_angles):
                # Varying radius with multiple sine waves
                radius_variation = (
                    base_size * (1 + 
                        np.sin(angle * frequency_variation) * deformation_intensity +
                        np.cos(angle * (frequency_variation * 1.5)) * (deformation_intensity * 0.5)
                    )
                )
                
                x.append(center_x + radius_variation * np.cos(angle))
                y.append(center_y + radius_variation * np.sin(angle))
            
            # Close the shape
            x.append(x[0])
            y.append(y[0])
            
            # Plot the sculptural form
            self.plot_widget.plot(x, y, 
                                   pen={'color': color, 'width': 2},
                                   fill=True, 
                                   brush=pg.mkBrush(color=color, alpha=0.2))
            
            # Add internal structural lines
            for i in range(len(x) - 1):
                mid_x = (x[i] + center_x) / 2
                mid_y = (y[i] + center_y) / 2
                self.plot_widget.plot(
                    [center_x, x[i]], 
                    [center_y, y[i]], 
                    pen={'color': color, 'width': 1, 'style': Qt.PenStyle.DotLine}
                )
        
        # Create multiple interactive sculptures
        create_sound_sculpture(
            center_x=0.5, 
            center_y=0.5, 
            base_size=0.3, 
            complexity=rms_normalized, 
            color=base_colors[0]
        )
        
        create_sound_sculpture(
            center_x=0.5, 
            center_y=0.5, 
            base_size=0.2, 
            complexity=peak_normalized, 
            color=base_colors[1]
        )
        
        # Set title and ensure consistent display
        self.plot_widget.setTitle('Artistic Sound Sculpture', color='#2c3e50')
        self.plot_widget.getPlotItem().setAspectLocked(True, ratio=1)
        self.plot_widget.getPlotItem().getViewBox().setRange(
            xRange=[0, 1], 
            yRange=[0, 1]
        )

    def letter_field_visualizer(self, audio_data):
        """
        Create a 20x20 grid of letters that change based on audio intensity
        """
        self.plot_widget.clear()
        
        # Compute audio characteristics
        rms = np.sqrt(np.mean(audio_data**2))
        peak = np.max(np.abs(audio_data))
        
        # Logarithmic normalization for sensitivity
        rms_normalized = min(np.log1p(rms * 20), 1)
        peak_normalized = min(np.log1p(peak * 25), 1)
        
        # Letter sets with different visual weights
        letter_sets = [
            'ABCDEFGHIJKLMNOPQRSTUVWXYZ',  # Uppercase
            'abcdefghijklmnopqrstuvwxyz',  # Lowercase
            '0123456789',                  # Numbers
            '!@#$%^&*()_+=-[]{}|;:,.<>?'   # Symbols
        ]
        
        # Color palette (silver to blue tones)
        color_palette = [
            '#7f8c8d',   # Soft gray
            '#95a5a6',   # Silver gray
            '#bdc3c7',   # Light silver
            '#2980b9',   # Deep blue
            '#3498db',   # Bright blue
            '#34495e',   # Dark blue-gray
            '#2c3e50',   # Deep blue-gray
        ]
        
        def generate_letter_grid(intensity):
            grid = []
            for y in range(20):
                row = []
                for x in range(20):
                    # Randomize letter selection based on intensity
                    letter_set = letter_sets[int(intensity * (len(letter_sets) - 1))]
                    letter = random.choice(letter_set)
                    
                    # Vary letter size and color based on intensity
                    font_size = 8 + int(intensity * 12)  # 8-20 point size
                    color = color_palette[int(x/5 + y/5) % len(color_palette)]
                    
                    row.append((letter, font_size, color))
                grid.append(row)
            return grid
        
        # Generate grids based on different audio characteristics
        rms_grid = generate_letter_grid(rms_normalized)
        peak_grid = generate_letter_grid(peak_normalized)
        
        # Combine grids
        combined_grid = []
        for y in range(20):
            row = []
            for x in range(20):
                # Alternate between RMS and Peak grid characteristics
                if (x + y) % 2 == 0:
                    row.append(rms_grid[y][x])
                else:
                    row.append(peak_grid[y][x])
            combined_grid.append(row)
        
        # Plot the letter field
        for y, row in enumerate(combined_grid):
            for x, (letter, font_size, color) in enumerate(row):
                text_item = pg.TextItem(
                    text=letter, 
                    color=color, 
                    anchor=(0.5, 0.5)
                )
                # Use QFont instead of pg.mkFont
                font = QFont()
                font.setPointSize(font_size)
                text_item.setFont(font)
                text_item.setPos(x/10, y/10)
                self.plot_widget.addItem(text_item)
        
        # Set title and ensure consistent display
        self.plot_widget.setTitle('Dynamic Letter Field', color='#2c3e50')
        self.plot_widget.getPlotItem().setAspectLocked(True, ratio=1)
        self.plot_widget.getPlotItem().getViewBox().setRange(
            xRange=[0, 2], 
            yRange=[0, 2]
        )

    def dynamic_circle_interaction_visualizer(self, audio_data):
        """
        Create a dynamic visualization of circles interacting 
        based on different audio frequency bands with enhanced dynamics
        """
        self.plot_widget.clear()
        
        # Compute frequency bands
        spectrum = np.abs(np.fft.rfft(audio_data))
        
        # Split spectrum into bass, mid, and treble bands
        bass_band = spectrum[:len(spectrum)//3]
        mid_band = spectrum[len(spectrum)//3:2*len(spectrum)//3]
        treble_band = spectrum[2*len(spectrum)//3:]
        
        # Normalize band energies
        bass_energy = min(np.mean(bass_band) * 5, 1)
        mid_energy = min(np.mean(mid_band) * 5, 1)
        treble_energy = min(np.mean(treble_band) * 5, 1)
        
        # Color palette (silver to blue tones)
        color_palette = [
            '#7f8c8d',   # Soft gray
            '#95a5a6',   # Silver gray
            '#bdc3c7',   # Light silver
            '#2980b9',   # Deep blue
            '#3498db',   # Bright blue
            '#34495e',   # Dark blue-gray
            '#2c3e50',   # Deep blue-gray
        ]
        
        # Store circle dynamics as instance attribute if not exists
        if not hasattr(self, '_circle_dynamics'):
            self._circle_dynamics = {
                'bass': {
                    'center_x': 0.3, 
                    'center_y': 0.5,
                    'velocity_x': 0,
                    'velocity_y': 0,
                    'jelly_phase': 0
                },
                'mid': {
                    'center_x': 0.5, 
                    'center_y': 0.5,
                    'velocity_x': 0,
                    'velocity_y': 0,
                    'jelly_phase': 0
                },
                'treble': {
                    'center_x': 0.7, 
                    'center_y': 0.5,
                    'velocity_x': 0,
                    'velocity_y': 0,
                    'jelly_phase': 0
                }
            }
        
        # Dynamic circle creation with movement and jelly effect
        def create_dynamic_circle(circle_type, base_radius, energy, color):
            dynamics = self._circle_dynamics[circle_type]
            
            # Add random velocity (controlled by audio energy)
            dynamics['velocity_x'] += (random.random() - 0.5) * energy * 0.02
            dynamics['velocity_y'] += (random.random() - 0.5) * energy * 0.02
            
            # Damping to prevent excessive movement
            dynamics['velocity_x'] *= 0.95
            dynamics['velocity_y'] *= 0.95
            
            # Update position
            dynamics['center_x'] += dynamics['velocity_x']
            dynamics['center_y'] += dynamics['velocity_y']
            
            # Boundary constraints
            dynamics['center_x'] = max(0.1, min(dynamics['center_x'], 0.9))
            dynamics['center_y'] = max(0.1, min(dynamics['center_y'], 0.9))
            
            # Jelly-like oscillation
            dynamics['jelly_phase'] += energy * 0.5
            jelly_factor = 1 + np.sin(dynamics['jelly_phase']) * energy * 0.2
            
            # Dynamic radius with jelly effect
            radius = base_radius * (1 + energy) * jelly_factor
            
            # Create circle with pulsating effect
            theta = np.linspace(0, 2*np.pi, 100)
            
            # Add subtle deformation based on energy
            deformation = np.sin(theta * 3 + dynamics['jelly_phase']) * energy * 0.1
            
            x = dynamics['center_x'] + (radius + deformation) * np.cos(theta)
            y = dynamics['center_y'] + (radius + deformation) * np.sin(theta)
            
            # Plot the circle
            self.plot_widget.plot(x, y, 
                                   pen={'color': color, 'width': 2},
                                   fill=True, 
                                   brush=pg.mkBrush(color=color, alpha=0.3))
            
            # Add connecting lines to center
            for angle in np.linspace(0, 2*np.pi, 6, endpoint=False):
                line_x = [dynamics['center_x'], 
                          dynamics['center_x'] + radius * np.cos(angle)]
                line_y = [dynamics['center_y'], 
                          dynamics['center_y'] + radius * np.sin(angle)]
                self.plot_widget.plot(line_x, line_y, 
                                       pen={'color': color, 'width': 1, 'style': Qt.PenStyle.DotLine})
        
        # Create circles with dynamic positioning
        create_dynamic_circle(
            circle_type='bass',
            base_radius=0.2, 
            energy=bass_energy, 
            color=color_palette[3]  # Deep blue
        )
        
        create_dynamic_circle(
            circle_type='mid',
            base_radius=0.15, 
            energy=mid_energy, 
            color=color_palette[4]  # Bright blue
        )
        
        create_dynamic_circle(
            circle_type='treble',
            base_radius=0.1, 
            energy=treble_energy, 
            color=color_palette[5]  # Dark blue-gray
        )
        
        # Add interaction lines between circles
        interaction_colors = [color_palette[2], color_palette[6]]
        dynamics = self._circle_dynamics
        for i in range(2):
            self.plot_widget.plot(
                [dynamics['bass']['center_x'], 
                 dynamics['mid']['center_x'], 
                 dynamics['treble']['center_x']],  # X coordinates
                [dynamics['bass']['center_y'], 
                 dynamics['mid']['center_y'], 
                 dynamics['treble']['center_y']],  # Y coordinates
                pen={'color': interaction_colors[i], 'width': 1, 'style': Qt.PenStyle.DashLine}
            )
        
        # Set title and ensure consistent display
        self.plot_widget.setTitle('Dynamic Frequency Circles', color='#2c3e50')
        self.plot_widget.getPlotItem().setAspectLocked(True, ratio=1)
        self.plot_widget.getPlotItem().getViewBox().setRange(
            xRange=[0, 1], 
            yRange=[0, 1]
        )

    def auto_change_visualization(self):
        """
        Automatically change visualization only if auto-change is enabled
        """
        if self.is_playing and self.auto_change_enabled:
            self.current_viz_index = (self.current_viz_index + 1) % len(self.viz_modes)
            
            # Optional: Highlight the current visualization button
            for btn in self.viz_buttons:
                btn.setStyleSheet("""
                    QPushButton {
                        background-color: rgb(236, 240, 241);
                        color: rgb(44, 62, 80);
                        border: 1px solid rgb(189, 195, 199);
                        border-radius: 4px;
                        padding: 5px;
                        font-size: 10px;
                    }
                    QPushButton:hover {
                        background-color: rgb(213, 219, 219);
                    }
                    QPushButton:pressed {
                        background-color: rgb(192, 198, 204);
                    }
                """)
            
            # Highlight the current visualization button
            self.viz_buttons[self.current_viz_index].setStyleSheet("""
                QPushButton {
                    background-color: rgb(52, 152, 219);
                    color: white;
                    border: 1px solid rgb(41, 128, 185);
                    border-radius: 4px;
                    padding: 5px;
                    font-size: 10px;
                }
            """)

    def three_channel_3d_disco_light_sequencer(self, audio_data):
        """
        Draw three filled squares in RGB, representing a 3 channel 1980s disco light sequencer.
        """
        self.plot_widget.clear()
        
        # Define the colors for the three channels
        colors = [QColor('#FF5733'), QColor('#33FF57'), QColor('#3357FF')]  # Bright Red, Green, Blue
        positions = [0.0, 0.5, 1.0]  # X positions for the squares
        square_size = 0.45  # Size of the squares (width and height)

        # Draw the casing
        casing_color = QColor('#2c3e50')  # Dark grey color for the casing
        casing = pg.BarGraphItem(x0=-0.1, y0=0.1, x1=1.1, y1=0.75, pen=casing_color, brush=casing_color)
        self.plot_widget.addItem(casing)

        # Compute the spectrum and split into bands
        spectrum = np.abs(np.fft.rfft(audio_data))
        bass_band = spectrum[:len(spectrum)//3]
        mid_band = spectrum[len(spectrum)//3:2*len(spectrum)//3]
        treble_band = spectrum[2*len(spectrum)//3:]

        # Adjust sensitivity: less for bass, more for mid and treble
        bass_energy = min(np.mean(bass_band) * 3, 1)  # Less sensitive (multiplier reduced)
        mid_energy = min(np.mean(mid_band) * 20, 1)    # More sensitive (multiplier increased)
        treble_energy = min(np.mean(treble_band) * 25, 1)  # More sensitive (multiplier increased)
        
        # Draw the squares with dynamic colors based on energy
        for i, (color, energy) in enumerate(zip(colors, [bass_energy, mid_energy, treble_energy])):
            # Adjust brightness based on energy
            brightness = 0.5 + 0.5 * energy  # Scale brightness from 0.5 to 1
            dynamic_color = pg.mkColor(color.red() * brightness, color.green() * brightness, color.blue() * brightness)  # Adjust color brightness
            
            square = pg.BarGraphItem(x0=positions[i] - square_size / 2, y0=0.25, 
                                     x1=positions[i] + square_size / 2, y1=square_size + 0.25, 
                                     pen=dynamic_color, brush=dynamic_color)
            self.plot_widget.addItem(square)

            # Add shadow effect to the square
            shadow_effect = QGraphicsDropShadowEffect()
            shadow_effect.setBlurRadius(10)
            shadow_effect.setColor(QColor(0, 0, 0, 150))  # Semi-transparent black
            shadow_effect.setOffset(2, 2)  # Offset for the shadow
            
            # Apply the shadow effect to the square
            square.setGraphicsEffect(shadow_effect)
        
        self.plot_widget.setTitle('3D Disco Light Sequencer', color='#2c3e50')
        self.plot_widget.getPlotItem().setAspectLocked(True)
        self.plot_widget.getPlotItem().getViewBox().setRange(
            xRange=[-0.1, 1.1], 
            yRange=[0, 0.75]
        )

    def toggle_disco_mode(self):
        """
        Toggle between normal visualization and dedicated Disco Light Sequencer
        """
        # Find the index of the Disco Light Sequencer in viz_modes
        disco_index = self.viz_modes.index(self.three_channel_3d_disco_light_sequencer)
        
        # Set current visualization to Disco Light Sequencer
        self.current_viz_index = disco_index
        
        # Optional: Add a visual indicator
        self.disco_btn.setStyleSheet("""
            background-color: #e74c3c;
            color: white;
            border: 2px solid #c0392b;
        """)
        
        # Reset the button style after a short delay
        QTimer.singleShot(1000, self.reset_disco_button_style)

    def reset_disco_button_style(self):
        """Reset the Disco button to its original style"""
        self.disco_btn.setStyleSheet("")

    def toggle_auto_change(self):
        """
        Toggle automatic visualization change
        """
        self.auto_change_enabled = not self.auto_change_enabled
        
        if self.auto_change_enabled:
            # Start the timer
            self.viz_change_timer.start(30000)  # 30 seconds
            
            # Visual feedback
            self.auto_change_btn.setStyleSheet("""
                QPushButton {
                    background-color: rgb(46, 204, 113);
                    color: white;
                    border: 2px solid rgb(39, 174, 96);
                }
            """)
        else:
            # Stop the timer
            self.viz_change_timer.stop()
            
            # Reset button style
            self.auto_change_btn.setStyleSheet("")

    def spirograph_visualizer(self, audio_data):
        """
        Create a Spirograph visualization based on audio characteristics
        with effects similar to the Original Spirograph Design Set.
        """
        self.plot_widget.clear()
        
        # Compute audio characteristics
        rms = np.sqrt(np.mean(audio_data**2))
        peak = np.max(np.abs(audio_data))
        
        # Normalize for visualization
        rms_normalized = min(np.log1p(rms * 20), 1)
        peak_normalized = min(np.log1p(peak * 25), 1)
        
        # Parameters for the Spirograph
        num_points = 1000
        theta = np.linspace(0, 2 * np.pi, num_points)
        
        # Create multiple spirograph patterns
        for i in range(10):  # Create 10 overlapping spirographs
            r = 0.1 + 0.05 * (i + 1) * rms_normalized  # Varying radius based on RMS
            R = 0.2 + 0.1 * (i + 1) * peak_normalized  # Varying radius based on Peak
            d = 0.1  # Distance from the center
            
            # Vary the angular velocity for each wheel
            angular_velocity = (i + 1) * 0.5  # Different speeds for each wheel
            
            # Calculate the x and y coordinates for the spirograph
            x = (R - r) * np.cos(theta * angular_velocity + i * np.pi / 10) + d * np.cos((R - r) / r * theta)
            y = (R - r) * np.sin(theta * angular_velocity + i * np.pi / 10) - d * np.sin((R - r) / r * theta)
            
            # Plot the Spirograph with varying colors
            color = pg.mkColor(255 * (i / 10), 0, 255 * (1 - i / 10), 150)  # Gradient color with alpha
            self.plot_widget.plot(x, y, pen=color)  # Use the gradient color for each spirograph
        
        self.plot_widget.setTitle('Spirograph Visualization', color='#2c3e50')
        
        # Ensure consistent display
        self.plot_widget.getPlotItem().setAspectLocked(True)
        self.plot_widget.getPlotItem().getViewBox().setRange(
            xRange=[-1, 1], 
            yRange=[-1, 1]
        )

    def emoji_visualizer(self, audio_data):
        """
        Create an emoji visualization that changes based on audio characteristics.
        """
        self.plot_widget.clear()
        
        # Compute audio characteristics
        rms = np.sqrt(np.mean(audio_data**2))
        peak = np.max(np.abs(audio_data))
        
        # Normalize for visualization
        rms_normalized = min(np.log1p(rms * 20), 1)
        peak_normalized = min(np.log1p(peak * 25), 1)
        
        # Define emojis based on mood
        emojis = {
            'happy': '😊',
            'sad': '😢',
            'angry': '😠',
            'surprised': '😲',
            'neutral': '😐'
        }
        
        # Determine mood based on audio characteristics
        mood = 'neutral'
        if rms_normalized > 0.7:
            mood = 'angry'  # High energy
        elif peak_normalized > 0.5:
            mood = 'happy'  # High peaks
        elif rms_normalized < 0.3:
            mood = 'surprised'  # Low energy
        
        # Display the selected emojis
        emoji_texts = [emojis[mood]] * 3  # Repeat the selected emoji three times
        positions = [-0.8, 0, 0.8]  # X positions for three emojis
        
        # Calculate dynamic font size based on audio characteristics
        base_font_size = 40 + (rms_normalized * 110)  # Scale between 40 and 150
        
        for i, emoji_text in enumerate(emoji_texts):
            emoji_item = pg.TextItem(text=emoji_text, color='#2c3e50', anchor=(0.5, 0.5))
            emoji_item.setFont(QFont("Arial", int(base_font_size)))  # Use dynamic font size
            
            # Add a shadow effect for better styling
            shadow_effect = QGraphicsDropShadowEffect()
            shadow_effect.setBlurRadius(10)
            shadow_effect.setColor(QColor(0, 0, 0, 100))  # Semi-transparent black
            shadow_effect.setOffset(2, 2)
            emoji_item.setGraphicsEffect(shadow_effect)
            
            emoji_item.setPos(positions[i], 0)  # Center the emojis in a row
            self.plot_widget.addItem(emoji_item)
        
        self.plot_widget.setTitle('Emoji Mood Visualization', color='#2c3e50')
        
        # Ensure consistent display
        self.plot_widget.getPlotItem().setAspectLocked(True)
        self.plot_widget.getPlotItem().getViewBox().setRange(
            xRange=[-1, 1], 
            yRange=[-1, 1]
        )

    def rainbow_wave_visualizer(self, audio_data):
        """
        Create a rainbow effect visualization using sinusoidal waves
        that respond to the beat of the audio with enhanced artistic effects
        and a rotating effect between -90 and 90 degrees.
        """
        self.plot_widget.clear()
        
        # Compute audio characteristics
        rms = np.sqrt(np.mean(audio_data**2))
        frequency = np.fft.rfftfreq(len(audio_data), d=1/self.sample_rate)
        
        # Normalize RMS for wave amplitude
        amplitude = min(np.log1p(rms * 20), 1)
        
        # Create sinusoidal waves with varying colors and artistic effects
        num_waves = self.current_wave_count  # Use the current wave count from the slider
        x = np.linspace(0, 2 * np.pi, 1000)
        
        # Get the current time for rotation effect
        current_time = time.time()
        rotation_speed = 1  # Adjust this value to change rotation speed
        
        for i in range(num_waves):
            # Vary frequency, phase, and amplitude for each wave
            frequency_factor = 1 + i * 0.3  # Adjust frequency factor for more variation
            phase_shift = i * (np.pi / 10)  # Smaller phase shift for smoother transitions
            wave_amplitude = amplitude * (0.5 + 0.5 * np.sin(i * 0.5))  # Vary amplitude dynamically
            
            # Calculate rotation angle based on time, oscillating between -90 and 90 degrees
            rotation_angle = np.sin(current_time * rotation_speed + i * (np.pi / num_waves)) * (np.pi / 2)  # -90 to 90 degrees in radians
            y = wave_amplitude * np.sin(frequency_factor * x + phase_shift + rotation_angle)  # Apply rotation
            
            # Create a color gradient based on the wave index
            color = pg.mkColor(255 * (i / num_waves), 0, 255 * (1 - i / num_waves), 150 + 100 * np.sin(i * 0.2))  # Add transparency variation
            self.plot_widget.plot(x, y + i * 0.1 - 0.75, pen=color)  # Offset each wave vertically for layering
        
        self.plot_widget.setTitle('Artistic Rainbow Wave Visualization', color='#2c3e50')
        
        # Ensure consistent display
        self.plot_widget.getPlotItem().setAspectLocked(False)
        self.plot_widget.getPlotItem().getViewBox().setRange(
            xRange=[0, 2 * np.pi], 
            yRange=[-1.1, num_waves * 0.1 - 0.5]
        )

    def update_wave_count(self, value):
        self.current_wave_count = value  # Store the current wave count
        self.wave_scale_label.setText(f"Number of Waves: {value}")  # Update scale label

def main():
    app = QApplication(sys.argv)
    player = AdvancedMP3Player()
    player.show()
    sys.exit(app.exec())

if __name__ == "__main__":
    main()
